// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// History:
// Original
// CRx5 jcv 10-Oct-08 Enhanced versioning management
// SPRx4 jcv 17-OCt-08 stsp file should reflect Stimes with 6 precision  
// CRx7 jcv 19-Jan-09 Enhanced tracing RTX creation 
// SPRx15 jcv 22-May-09 additional passes verbosity
// CRx18b jcv 10-Jul-09 Addition of abstract functions from PacketConsumer
// SPRx24 jcv 9-Feb-10 Additional trace information
// SPRx26 jcv 15-Feb-10 Spike enhancements
// SPRx28 jcv 16-Feb-10 Activation of UTC/ERT columns update after fitting 
// SPRx30 jcv/eo 11-Mar-10 Solaris compilation
// SPRx31 jcv 24-Mar-2010 Added moveup function for new columns
// CRx22  jcv 10-Nov-2011 cleanup may not happen in case analyses is needed 

#include "STSPProcessor.h"
#include "CalOalUtility.h"
#include "RTX.h"

#include "param.h"

#include<sstream>
#include<fstream>

// File for STSP output;
ofstream file; 

///////////////////////
//
//
//
//
STSPProcessor::STSPProcessor(int rev, int ver) : Revolution(rev), Version(ver) {

  stringstream s; 
  s.width(4); s.fill('0');
  s << rev;
  revString = s.str();

  // CR x5 Enhanced version management
  stringstream sver; 
  sver.width(4); sver.fill('0');
  sver << ver;
  verString = sver.str();

  // CRx5 Enhanced version management
  //stringstream fname;  fname << "RTX_" << revString << ".stsp"; 
  stringstream fname;  fname << "RTX_" << revString << "_" << verString << ".stsp"; 
  file.open(fname.str().c_str());

//  refXMCSCorrection = STime("2000-01-30T12:00:00.0");
  refXMCSCorrection = STime("2002-01-22T12:00:00.0");

  // Have not received a frame yet
  haveFrame = false;
  
  // OBT reset detected
  hasOBTreset = false;

}

// CRx18b
bool STSPProcessor::willacceptPacket(RawTM::XMM::PacketHeader &hdr, RawTM::XMM::DataFieldHeader &dfh, unsigned char rawSID) {

 // This is here just because it was an abstract, and needs to be defined
 errstr << "Called obsolete function. Will accept all." 
      << Sas::message(Msg::App, Msg::Verbose);

 if(0) {
   std::cerr << dfh.TM.PacketSubtype << std::endl;
   std::cerr << hdr.PacketID.APID << std::endl;
   std::cerr << rawSID << std::endl;
 }

 return true;

}

bool STSPProcessor::hasTimePacket(RawTM::XMM::Packet &pkt) {

 // This is here just because it was an abstract, and needs to be defined
 errstr << "Called obsolete function. Will accept all."
      << Sas::message(Msg::App, Msg::Verbose);

 if(0) {
   std::cerr << pkt.Header.PacketID.APID << std::endl;
 }

 return false;

}
// END CRx18b

///////////////////////
//
//
//
//
void STSPProcessor::processFrame(RawTM::NCTRS::TMDU &frm) {
  
  currentFrame = frm;

  // Unwrap MFC
  MFC = unwrapMFC(frm.Frame.MFC, frm.Header.Time);
  
  // if (VFC % 16) == 0, STSP generation triggered
  if ((frm.Frame.VFC % 16) == 0) {
    
    // Received a frame
    haveFrame = true;

    // Record time and MFC
    MFCVC0 = MFC;
    TimeVC0 = Utils::cds2STime(frm.Header.Time);
    
  }
  
  return;

}

///////////////////////
//
//
//
//
bool STSPProcessor::acceptFrame(RawTM::NCTRS::FrameHeader &hdr) {
  
  if(0) {
    std::cerr << hdr.GSID << std::endl;
  }
  
  return true;

}

///////////////////////
//
//
//
//
void STSPProcessor::processPacket(RawTM::XMM::Packet &pkt, STime &time) {


  if (0) {
    cout << time.toString(6);
  }

  // Only generate an STSP record if we have received the
  // corresponding frame
  if (haveFrame) {

    // Current Frame Time
    STime currentFrameTime = Utils::cds2STime(currentFrame.Header.Time);
    
    // Trigger time, corrected for delays 
    
    double xmcsOBDHDelay = delayHandler.getXMCSOBDHdelay(currentFrameTime);
    double xmcsGSDelay   = delayHandler.getXMCSGSdelay(currentFrameTime, currentFrame.Header.GSID);
    
    double actualOBDHDelay = delayHandler.getActualOBDHdelay(currentFrameTime);
    double actualGSDelay   = delayHandler.getActualGSdelay(currentFrameTime, currentFrame.Header.GSID);
    
    
    // Before 2002-01-22T12:00:00Z, the XMCS overwrote the ERT. Correct for this
    // by reversing the changes
    
    // Trigger time, corrected for XMCS overwrite
    STime correctedXMCSTriggerTime;
    double signalDelay;
	
    // Calculate signal delay from ERT + XMCS correction (since the XMCS would have 
	// subtracted the GS Delay from the ERT
    if (TimeVC0 < refXMCSCorrection) {
	
//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " START " << TimeVC0.toString(6) 
//	<< std::endl;
	
	STime PropDelayCorrected = TimeVC0 + xmcsGSDelay;
    signalDelay = delayHandler.getSignalDelay(PropDelayCorrected, currentFrame.Header.GSID);
	
	correctedXMCSTriggerTime = TimeVC0 + signalDelay;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " PLUS PROP " << correctedXMCSTriggerTime.toString(6) 
//	<< " ( " << signalDelay << " ( "
//	<< std::endl;

	correctedXMCSTriggerTime = correctedXMCSTriggerTime + xmcsGSDelay;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " PLUS XMCS " << correctedXMCSTriggerTime.toString(6) 
//	<< " ( " << xmcsGSDelay << " ( "
//	<< std::endl;

	correctedXMCSTriggerTime = correctedXMCSTriggerTime + xmcsOBDHDelay;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " PLUS OBDH " << correctedXMCSTriggerTime.toString(6) 
//	<< " ( " << xmcsOBDHDelay << " ( "
//	<< std::endl;

    }
    else {
      signalDelay = delayHandler.getSignalDelay(TimeVC0, currentFrame.Header.GSID);	
	  correctedXMCSTriggerTime = TimeVC0;
    }
    
    double signalDelay2 = delayHandler.getSignalDelay(correctedXMCSTriggerTime - signalDelay, currentFrame.Header.GSID);
    
    // Now add the various delays
    STime correctedTriggerTime = correctedXMCSTriggerTime;
	
	correctedTriggerTime = correctedTriggerTime - signalDelay2;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " MINUS PROP " << correctedTriggerTime.toString(6) 
//	<< " ( " << signalDelay2 << " ( "
//	<< std::endl;

	correctedTriggerTime = correctedTriggerTime - actualGSDelay;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " MINUS XMCS " << correctedTriggerTime.toString(6) 
//	<< " ( " << actualGSDelay << " ( "
//	<< std::endl;

	correctedTriggerTime = correctedTriggerTime - actualOBDHDelay;

//	std::cout << "DEBUG " << pkt.DFHeader.Time.Coarse
//	<< " MINUS OBDH " << correctedTriggerTime.toString(6) 
//	<< " ( " << actualOBDHDelay << " ( "
//	<< std::endl;
	    	
    // Create a STSP record
    Utils::CUCTime pkTime;
    pkTime.TCoarse = pkt.DFHeader.Time.Coarse;
    pkTime.TFine = pkt.DFHeader.Time.Fine;
    
    file.precision(16);
    file << pkt.DFHeader.Time.Coarse
	 << "\t"
	 << pkt.DFHeader.Time.Fine
	 << "\t"
	 << ((double)pkt.DFHeader.Time.Fine / 65535.0) + (double)pkt.DFHeader.Time.Coarse
	 << "\t"
         // SPRx4 
	 //<< TimeVC0
	 << TimeVC0.toString(6)
	 << "\t"
         // SPRx4 
	 //<< currentFrameTime
	 << currentFrameTime.toString(6)
	 << "\t"
	 << signalDelay
	 << "\t"
	 << signalDelay2
	 << "\t"
	 << (int)currentFrame.Header.GSID
	 << "\t";
    
    file.precision(6);
    file << actualGSDelay
	 << "\t"
	 << actualOBDHDelay
	 << "\t"
	 << correctedTriggerTime.toString(6)
	 << "\t";
    
    file.precision(16);
    file << Utils::STime2MJD2000(correctedTriggerTime)
	 << std::endl;
    
    STSP stspRecord = STSP(pkTime,
			   (int)currentFrame.Header.GSID,
			   actualGSDelay,
			   actualOBDHDelay,
			   signalDelay,
			   TimeVC0,
			   correctedXMCSTriggerTime,
			   correctedTriggerTime,
			   currentFrameTime,
			   MFCVC0,
			   MFC);
    
    // Add it to the list
    packetList.push_back(stspRecord);

  }
  
  return;

}

///////////////////////////////////////
//
//
//
//
bool STSPProcessor::acceptPacket(RawTM::XMM::PacketHeader &hdr, RawTM::XMM::DataFieldHeader &dfh) {

  if (hdr.PacketID.APID == 0) {
    return true;
  }
  else {
    return false;
  }
  
  if(0) {
    std::cerr << dfh.TM.PacketSubtype << std::endl;
  }
  
  return true;
}


//////////////////////////////
//
//
//
//
int STSPProcessor::unwrapMFC(unsigned char mfc, Utils::CDSTime time) {
  
  static bool isInitMFCUnwrap = false;
  static int wrap;
  static STime timeSave;
  static int mfcSave;
  
  if (!isInitMFCUnwrap) {
    isInitMFCUnwrap = true;
    wrap = 0;
    
    timeSave = Utils::cds2STime(time);
    mfcSave = mfc;
    
  }
  
  STime frameTime = Utils::cds2STime(time);
  
  // Get the delta time between this frame and the last processed
  double timeDelta = frameTime.modifiedJulianDay().full() - timeSave.modifiedJulianDay().full();
  
  // Convert the delta time into a pseudo-Framecount
  timeDelta *= 86400.0 / (0.1015625 / ( 0.9999955 - 0.000003 * ( frameTime.modifiedJulianDay().full() / 1000.0 )));
  
  // Calculate the predicted MFC
  // SPRx30 Solaris compilation
#ifdef SunOS

  int mfcPred = mfcSave + rint(timeDelta);

#else

  int mfcPred = mfcSave + (int)round(timeDelta);

#endif

  // Why??
  if (mfcPred < 0) {
    wrap--;
  }
  
  // Why??
  int i = mfcPred % 256;
  
  // Why??
  if (i < 0) {
    i += 256;
  }
  
  // If predicted MFC not equal to actual MFC something wrong
  if (i != mfc) {
    errstr << "Inconsistent frame arrival time: " << frameTime
	   << Sas::warning("WrongMFC");  
  }
  
  // Calculate new wrap value
  // Why??
  if (mfcPred > 256) {
    wrap += mfcPred/256;
  }
  
  // Save frame details
  mfcSave = mfc;
  timeSave = frameTime;
  
  // Return unwrapped MFC
  return (mfc + (wrap * 256));
  
}


///////////////////////
//
//
//
//
void STSPProcessor::createRTX() {

  int nPass = 0;
  unsigned int lastGS = 0;
 
  double OBT = 0.0;
  double ERT = 0.0;
  Utils::CDSTime cdsT;

  double lastOBT = -1.0;
  double lastERT = 0.0;
  double postResetOBT = 0.0;
  double correctedOBT = 0.0;

  int currentRow = 0;

  // Set the file creator name
  dataSetServer->client( "stspproc" );

  // NEW CR x7
  //std::cout << Start << ":" << End << ":" << Revolution << ":" << Version << ":" << packetList.size() << std::endl;
  errstr << "create RTX in action " << Sas::message(Msg::App, Msg::Sparse);
  std::cout << "=== stspproc RTX creation process beginning ===" << std::endl;
  std::cout << "From: " << Start << std::endl;
  std::cout << "To: " << End << std::endl;
  std::cout << "Revolution: " << Revolution <<  " Version : " << Version << std::endl;
  std::cout << "STSP packet List size (n of OBT-ERT pairs): " << packetList.size() << std::endl;

  std::cout << "RTX: new instance called  ===" << std::endl;
  rtxFile = new RTX(Start, End, Revolution, Version, packetList.size()); 

  std::cout << "RTX: now will loop over OBT-ERT pairs  ===" << std::endl;

  for (unsigned int i = 0; i < packetList.size(); i++) {
    
    OBT = ((double)packetList[i].OBT.TCoarse + 
	   (double)packetList[i].OBT.TFine / 65535.0);

    ERT = Utils::STime2MJD2000(packetList[i].correctedSTSPTime) * 86400.0;

	double obtDiff;
	
	if (hasOBTreset) {
		obtDiff = correctedOBT + (OBT - postResetOBT) - lastOBT; 
	}
	else {
		obtDiff = OBT - lastOBT;
	}
	
	if (obtDiff < 0.0) {
	
		if (-obtDiff > 0.9 * 65536.0 * 65536.0) {

    	  errstr << "OBT wrap" 
	    	 << " ERT: " 
	    	 << packetList[i].currentFrameTime 
	    	 << " ERT (MJD2000): " 
	    	 << Utils::STime2MJD2000(packetList[i].currentFrameTime)
	    	 << " OBT: " 
	    	 << OBT
	    	 << " OBT diff: " 
	    	 << obtDiff
	    	 << Sas::warning("OBTwrap");
		
		}
		else {
		
		  postResetOBT = OBT;
		  correctedOBT = lastOBT + (ERT - lastERT);
		  		  
		  cout << postResetOBT << " " << correctedOBT << endl;

    	  errstr << "OBT reset" 
	    	 << " ERT: " 
	    	 << packetList[i].currentFrameTime 
	    	 << " ERT (MJD2000): " 
	    	 << Utils::STime2MJD2000(packetList[i].currentFrameTime)
	    	 << " OBT: " 
	    	 << OBT
	    	 << " OBT diff: " 
	    	 << obtDiff
	    	 << Sas::warning("OBTreset");

		  hasOBTreset = true;
			
		}

    }
	
	if (hasOBTreset) {
		OBT = correctedOBT + (OBT - postResetOBT);
        packetList[i].OBT.TCoarse = static_cast<int>(OBT);
        packetList[i].OBT.TFine = static_cast<int>((OBT - packetList[i].OBT.TCoarse) * 65535.0);
		cout << OBT <<  " " << packetList[i].OBT.TCoarse << " " << packetList[i].OBT.TFine << endl;
		
	}
	

    lastOBT = OBT;
    lastERT = ERT;

    // Filter bad STSPs
    if ( (i > 1) && ( packetList[i].uncorrectedSTSPTime == packetList[i - 1].uncorrectedSTSPTime)) {

      errstr << "(VFC%16=0) Lost:" 
	     << " ERT: " 
	     << packetList[i].currentFrameTime 
	     << " ERT (MJD2000): " 
	     << Utils::STime2MJD2000(packetList[i].currentFrameTime)
	     << " OBT: " 
	     << OBT
	     << Sas::warning("VFCLost");

      continue;
    }

    if ((packetList[i].currentFrameTime - packetList[i].correctedXMCSSTSPTime) > 4.0) {
      
      errstr << "LOG: (VFC%16==0) Lost && STSP Lost " 
	     << " ERT: " 
	     << packetList[i].currentFrameTime 
	     << " ERT (MJD2000): " 
	     << Utils::STime2MJD2000(packetList[i].currentFrameTime)
	     << " OBT: " 
	     << OBT 
	     << Sas::warning("STSPLost");

      continue;
    }

    // If we have switched ground stations
    if (lastGS != packetList[i].GSID) {

      // If this is not the first pass in the dataset
      if (lastGS != 0) {
	
	// Increment the number of passes
	nPass++;

      }

    }

    // Get the STSP in CDS format
    cdsT = Utils::STime2cds(packetList[i].correctedSTSPTime);

    // Get the GS ID in a string
    stringstream s; s << packetList[i].GSID;

    // Write the data to the file
    rtxFile->frameData[currentRow]   = currentRow + 1;
    rtxFile->obtcData[currentRow]    = packetList[i].OBT.TCoarse;
    rtxFile->obtfData[currentRow]    = packetList[i].OBT.TFine;
    rtxFile->utcdayData[currentRow]  = cdsT.Days;
    rtxFile->utcmilData[currentRow]  = cdsT.Milliseconds;
    rtxFile->utcmicData[currentRow]  = cdsT.Microseconds;
    rtxFile->utcflagData[currentRow] = 0; // good
    CalOalUtility::stringColumnRowSet(rtxFile->ertColD, currentRow, packetList[i].correctedSTSPTime.toString(6));
    CalOalUtility::stringColumnRowSet(rtxFile->gsidColD, currentRow, s.str());
    CalOalUtility::stringColumnRowSet(rtxFile->ertrawColD, currentRow, packetList[i].uncorrectedSTSPTime.toString(6));
    rtxFile->utcreaData[currentRow]  = ERT;
    rtxFile->obtrData[currentRow]    = OBT;
    rtxFile->passData[currentRow]    = nPass;
    rtxFile->gsData[currentRow]      = packetList[i].GSID;
    rtxFile->predERTData[currentRow] = 0.0;
    rtxFile->residData[currentRow]   = 0.0;

    // Remember last GS
    lastGS = packetList[i].GSID;

    // Increment the current row
    currentRow++;

  }


  // Delete extra rows
  rtxFile->tab->deleteRows(currentRow,packetList.size() - currentRow);
  
  // Close the dataset  
  rtxFile->close();
  
  // NEW CR x7
  std::cout << "=== stspproc RTX file creation process ended ===" << std::endl;
}

///////////////////////
//
//
//
//
void STSPProcessor::correctSTSPs(ostream & os) {
  
  map<int, Pass *>passList;

  unsigned int i;
  unsigned int nSymCorrections   = 0;
  unsigned int nSpikeCorrections = 0;

  openParameters("stspproc");
  int fitPoints = intParameter("passfitpoints");
  int passDelay = intParameter("postpassdelay");
  double maxDelta = realParameter("maxdelta");
  double symThres = realParameter("symthres");
  int searchWindow = intParameter("jumpsearchwindow");
  bool searchJumps = booleanParameter("searchjumps");
  double maxResidual =  realParameter("maxresid");
  double maxTimeJump =  realParameter("maxtimejump");
  bool correctAll = booleanParameter("correctall");
  bool markBad = booleanParameter("markbad");

  // NEW CRx22
  bool keepAncilliary = booleanParameter("keepancilliary");
    
  // NEW CR x7
  os << "Opening RTX file for fit, correctSTSP in action  ===" << std::endl;
  std::cout << "runlog === Opening RTX file for fit, correctSTSP in action ===" << std::endl;

  rtxFile->open();

  // Generate fit for the revolution
  Pass initialFit = rtxFile->fitRevolution();

  // SPRx24
  std::cout << "runlog === Revolution fitted ===" << std::endl;

  // NEW CR x7
  os << "Plot RMS beginning (ORIGINAL) correctSTSP in action ===" << std::endl;
  errstr << "Plot RMS beginning (ORIGINAL) correctSTSP in action ===" << std::endl;
  std::cout << "Plot RMS beginning (ORIGINAL) correctSTSP in action ===" << std::endl;

  stringstream f1; f1 << initialFit.getRMS();

  rtxFile->plotRevolution("Fit on raw data for rev " + revString + " RMS: " + f1.str(), "" ,"_ORIGINAL.ps");

  // NEW CR x8
  // Commented out 
  //rtxFile->plotInputsRevolution("Raw data rev " + revString , "" ,"_RAW.ps");

  // NEW CR x7
  os << "correctSTSP: Fitting passes begins ===" << std::endl;
  errstr << "correctSTSP: Fitting passes begins  ===" << std::endl;
  std::cout << "runlog " << std::endl;
  std::cout << "runlog ===== correctSTSP:  RTX fitting passes beginning ===" << std::endl;
  std::cout << "runlog " << std::endl;

  // Calculate fit for passes
  passList = rtxFile->fitPasses(false, false);

  // SPRx24
  os << " " << std::endl;
  os << "correctSTSP: Fitting passes ends ===" << std::endl;
  os << " " << std::endl;
  errstr << "correctSTSP: Fitting passes ends  ===" << std::endl;
  std::cout << "runlog correctSTSP:  RTX fitting passes ends ===" << std::endl;
  std::cout << " " << std::endl;
  std::cout << "runlog detailed Passes info in other log " << std::endl;

  int nPriorPasses = 0;
  int nPostPasses = 0;

  int row;
  int lastStation;

  // Output RMS for each pass fit
  for (i = 0; i < passList.size(); i++) {

    row = passList[i]->getStart();
    STime PassStart = Utils::MJD20002STime(rtxFile->utcreaData[passList[i]->getStart()] / 86400.0);
    STime PassEnd   = Utils::MJD20002STime(rtxFile->utcreaData[passList[i]->getEnd()] / 86400.0);

    if (i > 0) {

      lastStation = rtxFile->gsData[ passList[i - 1]->getStart() ];

    }
    else {

      lastStation = 0;

    }

    os.precision(12);

    if (PassEnd < Start) {

      os << "->Pass: " 
	 << i
	 << " Begins with h/o from g/s "
	 << lastStation
	 << " to "
	 << passList[i]->getGSID()
	 << " Begins @ "
	 << PassStart
	 << " Ends @ " 
	 << PassEnd 
	 << std::endl;

      nPriorPasses++;

    }
    else if (PassStart > End) {

      os << "<-Pass: " 
	 << i
	 << " Begins with h/o from g/s  "
	 << lastStation
	 << " to "
	 << passList[i]->getGSID()
	 << " Begins @ "
	 << PassStart
	 << " Pass Ends @ " 
	 << PassEnd 
	 << std::endl;

      nPostPasses++;

    }
    else {

      os << "--Pass: " 
	 << i
	 << " Begins with h/o from g/s "
	 << lastStation
	 << " to "
	 << passList[i]->getGSID()
	 << " Begins @ "
	 << PassStart
	 << " Pass Ends @ " 
	 << PassEnd 
	 << std::endl;

      passList[i]->inRev = true;

    }

  }

  if (!nPostPasses || !nPriorPasses) {
    // No previous pass so abort here!

    errstr << "No pass detected outside revolution, Passes before: "
	   << nPriorPasses
	   << " Passes after: "
	   << nPostPasses
	   << Sas::fatal("noPass");
  }

  int thisPassNumber    = 0;
  int thisGSID          = 0;
  int lastGSID          = 0;
  int StationPoints     = 0;
  int lastStationPoints = 0;
  double predictedERT   = 0.0;  

  // NEW CR x7
  os << "correctSTSP: Calculating H/O delta begins ===" << std::endl;

  // Calculate H/O delta
  for (i = 0; i < rtxFile->tab->rows(); i++) {

    thisGSID       = rtxFile->gsData[i];
    thisPassNumber = rtxFile->passData[i];

    // Have we had a handover
    if (lastGSID != thisGSID) {
      lastStationPoints = StationPoints;
      StationPoints = 0;
      passList[thisPassNumber]->jump = 0.0;


      // NEw SPRx15
      if ( (lastStationPoints <  passDelay ) && (thisPassNumber >2 ))
        {
        os << "H/O detected, but unable to predict ERT and Delta for Pass: " 
	 << thisPassNumber-1
	 << " from "
	 << passList[thisPassNumber - 2]->getGSID()
	 << " to "
	 << passList[thisPassNumber - 1]->getGSID()
	 << " Pass Start @ "
	 << Utils::MJD20002STime(rtxFile->utcreaData[passList[thisPassNumber-1]->getStart()] / 86400.0)
	 << " Station Points were " << lastStationPoints 
         << " less than PassDelay  " <<  passDelay 
         << " .FitPoints are " <<  fitPoints
	 << std::endl;

         if (searchJumps) 
           os << "Forced to search Jumps" << std::endl;
            else
           os << "No additional jumps will be searched (re-run forcing it if required)" << std::endl;
         }
      // END SPRx15

    }
    else {
      StationPoints++;
    }

    // If we have had enough datapoints on this pass
    // AND we are not on the first pass
    // AND we have enough points for a good fit on last pass
    if ( (StationPoints == passDelay) &&
	 (thisPassNumber != 0) &&
	 (lastStationPoints > fitPoints) ) {

      // Calculate predicted ERT based on FIT for last pass      
      predictedERT = passList[thisPassNumber - 1]->getPredictedERT(rtxFile->obtrData[i]);
      passList[thisPassNumber]->jump = predictedERT - rtxFile->utcreaData[i];

      os << "++Pass: " 
	 << thisPassNumber
         // SPRx15
         << " h/o delta processed at passdelay STSP " << passDelay << "-th point" 
         // END SPRx15
         << std::endl;
         os << "     Previous Pass g/s "
	 << passList[thisPassNumber - 1]->getGSID()
	 << " This Pass g/s  "
	 << thisGSID
	 << " Pass Start @ "
	 << Utils::MJD20002STime(rtxFile->utcreaData[passList[thisPassNumber]->getStart()] / 86400.0)
	 << " (OBT: "
	 << rtxFile->obtrData[passList[thisPassNumber]->getStart()]
	 << ") " 
         // SPRx24
         << std::endl;
         os << "     Begins at file row " << i << " with h/o Delta @ " 
	 << Utils::MJD20002STime(rtxFile->utcreaData[i] / 86400.0)
	 << " (OBT: "
	 << rtxFile->obtrData[i]
	 << ") Predicted: " 
	 << predictedERT 
	 << " Observed: "
	 << rtxFile->utcreaData[i] 
	 << std::endl;
	 os << " Jump = Delta (predERT-UTC)  (* 1e6): "
	 << passList[thisPassNumber]->jump * 1e6
	 << std::endl;

    } 

    if ( (searchJumps) &&
	 (StationPoints < searchWindow) &&
	 (thisPassNumber != 0) &&
	 (lastStationPoints > fitPoints) ) {

      // Calculate predicted ERT based on FIT for last pass      
      // SPRx24
      os << "--Forced search of jumps and  station points " 
         << StationPoints << " less than searchWindow "
         << searchWindow << std::endl;
      os << "     Jumps: Found new jump Pass: "
	 << thisPassNumber
	 << " StationPoints: "
	 << StationPoints
	 << " OBT: "
	 << rtxFile->obtrData[i]
	 << " Delta: "
	 << rtxFile->utcreaData[i] - passList[thisPassNumber - 1]->getPredictedERT(rtxFile->obtrData[i])
	 << std::endl;
    } 

    lastGSID = thisGSID;

  }

  //
  //
  // Correct symmetric inter-pass jumps (at H/O)
  //
  //

  // First and last passes cannot be corrected by this method, but these
  // passes will fall outside the revolution since we request a time window 
  // either side of the rev.
  passList[0]->symmetricCorrection = 0.0;
  passList[passList.size() - 1]->symmetricCorrection = 0.0;

  double symmetry;

  // SPRx24
  os << " " << std::endl;
  os << "Now searching for jumps and correcting if requested and possible... " 
       << std::endl;
  os << " " << std::endl;

  // Detect jumps
  for (i = 1; i < passList.size() - 1; i++) {

    double time1 = rtxFile->utcreaData[ passList[i - 1]->getEnd() ];
    double time2 = rtxFile->utcreaData[ passList[i]->getStart() ];

    // SPRx24 Additional trace information
    //os << "Time delta (UTC_PreviousPassEnd - UTC_ThisPassStart) on H/O at Pass " << i << ": " << time2 - time1 << std::endl;
    os << " " << std::endl;
    os << "Pass: " << i 
       << " Begins with Jump : " 
       << passList[i]->jump 
       << " Time delta (UTC_PreviousPassEnd - UTC_ThisPassStart) on H/O : " 
       << time2 - time1 
       << std::endl;
       os << "  Pass within rev boundaries 1=true? " 
       << passList[i]->inRev
       << std::endl;

    // Only correct revolutions inside revolution
    if(correctAll || passList[i]->inRev) {

      // Jump on H/O
      if ( fabs(passList[i]->jump) > maxDelta) {

        // CR x7
	os << "--Jump: Catched Jump (predERT - UTC) " << passList[i]->jump << " larger than maxDelta " << maxDelta << std::endl;

	// Catch symmetric jump
	if ( ( fabs(passList[i + 1]->jump) > maxDelta ) &&
	     ( Utils::sign(passList[i]->jump) != Utils::sign(passList[i + 1]->jump)) ) {

          // JCV: Definition of symmetry index 
	  symmetry = (passList[i]->jump + passList[i + 1]->jump) / passList[i]->jump;

	  if ( fabs(symmetry) > symThres ) {

	    os << "--Jump: Catched Symmetric sign swap Jump on Pass, but Symmetry index Threshold " << symThres << " Broken. No correction. " 
	       << i
	       << ": Station: "
	       << rtxFile->gsData[passList[i]->getStart()]
	       << " to "
	       << rtxFile->gsData[passList[i + 1]->getStart()]
	       << " Begin Jump: "
	       << passList[i]->jump
	       << " End jump: " 
	       << passList[i + 1]->jump
	       << " Symmetry: "
	       << symmetry
	       << std::endl;

	    // No correction (yet)
	    passList[i]->symmetricCorrection = 0.0;

	  }
	  else {

	    // Catch big time jumps
	    if (time2 - time1 > maxTimeJump) {

	      os << "--Jump: Catched Symmetric sign swap (symmetry index within symthreshold) but UTCEnd-UTCStart is in abs values larger than " << maxTimeJump << ". No correction. "
		 << i
		 << ": Station: "
		 << rtxFile->gsData[passList[i]->getStart()]
		 << " to "
		 << rtxFile->gsData[passList[i + 1]->getStart()]
		 << " Begin Jump: "
		 << passList[i]->jump
		 << " End jump : " 
		 << passList[i + 1]->jump
		 << std::endl;

	      // No correction (yet)
	      passList[i]->symmetricCorrection = 0.0;

	    }
	    else {

	      passList[i]->symmetricCorrection = passList[i]->jump - ((passList[i]->jump + passList[i + 1]->jump) / 2.0);
	      passList[i + 1]->symmetricCorrection = 0.0;

              // SPRx24
	      os << "--Jump: Catched sign swap Symmetric Jump on Pass. "
                 << " and UTCEnd-UTCStart less than maxtimejump " << maxTimeJump
                 << " so Correction Possible.  " 
		 << std::endl;
              os << "  Details follows Pass: "
		 << i
		 << ": Station: "
		 << rtxFile->gsData[passList[i]->getStart()]
		 << " Ends with h/o to "
		 << rtxFile->gsData[passList[i + 1]->getStart()]
		 << " Begin Jump: "
		 << passList[i]->jump
		 << " End jump: " 
		 << passList[i + 1]->jump
                 << std::endl;
              os << "  Symmetry: "
		 << symmetry
		 << " Correction: "
		 << passList[i]->symmetricCorrection
		 << std::endl;

              // SPRx24
	      os << "--->remember: for this Pass GS  " 
	         << passList[i]->getGSID()
                 << " this GS " << thisGSID
                 << " , increasing Pass index to " << i+1 
		 << std::endl;


	      i++;
	      nSymCorrections++;

	    }

	  }

	}
	else {

	// CR x7 
	if ( fabs(passList[i + 1]->jump) > maxDelta ) 
	  os << "--Jump: Jump larger than maxDelta. But not change in sign. No correction." << std::endl;

	if ( Utils::sign(passList[i]->jump) != 
             Utils::sign(passList[i + 1]->jump) ) 
	  os << "--Jump: Change in sign but not larger than maxdelta. No correction." << std::endl;

	  os << "  Details follows: "   << i
	     << ": Station: "
	     << rtxFile->gsData[passList[i]->getStart()]
	     << " ending with h/o to "
	     << rtxFile->gsData[passList[i + 1]->getStart()]
	     << " Begin Jump: "
	     << passList[i]->jump
	     << " End Jump : " 
	     << passList[i + 1]->jump
	     << std::endl;

	  // No correction (yet)
	  passList[i]->symmetricCorrection = 0.0;

	}

      }
      else {
	// No jump, no correction
        // SPRx24 
	os << "Jump " << passList[i]->jump << " smaller than maxDelta " << maxDelta << std::endl;
	passList[i]->symmetricCorrection = 0.0;
      }
    }
    else {
      // Outside revolution
      // SPRx24
      os << "--Jump not corrected as pass outside rev boundaries " 
         << "(you can force by using correctall parameter) " 
         << std::endl;
      passList[i]->symmetricCorrection = 0.0;
    }
  }

  // Update the UTC Time
  // CR x7
  os << "If Symmetric Jumps were detected, correcting UTC" << std::endl;

  for (i = 0; i < rtxFile->tab->rows(); i++) {    
    rtxFile->utcreaData[i] = rtxFile->utcreaData[i] + passList[rtxFile->passData[i]]->symmetricCorrection;
  }

  // Cut the file to the revolution limits
  rtxFile->truncateFile();

  // Flush the data back to disk
  rtxFile->close();

  // NEW CR x7
  os << "RTX memory data flushed to disk  ===" << std::endl;
  std::cout << "runlog RTX memory data flushed to disk  === " << std::endl;

  // Re-open the file for the rest of the process
  rtxFile->open();

  // NEW CR x7
  // SPRx24
  os << " " << std::endl;
  os << "RTX disk data re-opened for whole revolution fit with Passes Jumps corrected " << std::endl;
  std::cout << " " << std::endl;
  std::cout << "RTX disk data re-opened for whole revolution fit with Passes Jumps corrected " << std::endl;

  // Calculate new fit
  // SPRx24
  //rtxFile->fitRevolution(true);
  Pass intermediateFit = rtxFile->fitRevolution(true);

  // SPRx24
  os << "...intermediate fit ended." << std::endl;
  std::cout << "...intermediate fit ended." << std::endl;

  if (markBad) {

        // CR x7 x26
        os << " " << std::endl;
        os << "Checking for spikes 100% to be marked as bad data. maxRes  " 
            << maxResidual << std::endl;
        std::cout << "runlog === Checking spikes 100% to be marked as bad." 
            << " maxRes " << maxResidual << std::endl;

	// Set spikes as bad data and set residual to zero
	for (i = 0; i < rtxFile->tab->rows(); i++) {    

    	   if ((fabs(rtxFile->residData[i]) > maxResidual * 100.0) &&
		(rtxFile->utcflagData[i] == 0)) {
    	        rtxFile->utcflagData[i] = 1;
    	        nSpikeCorrections++;
    	        }

	}

        // SPRx26
    	os << "Spikes 100%: Bad data: " << nSpikeCorrections << " points marked as bad" << std::endl;

        // CR x7
        os << "Again, fitting after bad spikes flagging  ===" << std::endl;
        std::cout << "=== Again, fitting after bad spikes flagging  ===" << std::endl;
	// Calculate new fit
	rtxFile->fitRevolution(true);

        // SPRx24 x26
        os << "Now re-flagging spikes 10% against latest fit" << std::endl;
        std::cout << "=== Now re-flagging spikes 10% against latest fit." 
            << " maxRes " << maxResidual << std::endl;

	// Set spikes as bad data and set residual to zero
	for (i = 0; i < rtxFile->tab->rows(); i++) {    

    	   if ((fabs(rtxFile->residData[i]) > maxResidual * 10.0) &&
		(rtxFile->utcflagData[i] == 0)) {
    	      rtxFile->utcflagData[i] = 1;
    	      nSpikeCorrections++;
    	      }

	}
        // SPRx26
    	os << "Spikes 10%: Bad data: " << nSpikeCorrections << " points marked as bad" << std::endl;

        // SPRx24 x26
        os << "Again, fitting after bad spikes flagging  ===" << std::endl;
        std::cout << "=== Again, fitting after bad spikes flagging." 
            << " maxRes " << maxResidual << std::endl;

	// Calculate new fit
	rtxFile->fitRevolution(true);

	// Set spikes as bad data and set residual to zero
	for (i = 0; i < rtxFile->tab->rows(); i++) {    

    	    if ((fabs(rtxFile->residData[i]) > maxResidual) &&
		(rtxFile->utcflagData[i] == 0)) {
    	        rtxFile->utcflagData[i] = 1;
    	        nSpikeCorrections++;
    	    }

	}

        // SPRx26
	//if (nSpikeCorrections > 0) {
    	os << "Bad data: " << nSpikeCorrections << " points marked as bad" << std::endl;
	//}

  }

  // SPRx24
  os << " " << std::endl;
  os << "FINAL whole revolution fit" << std::endl;
  std::cout << " " << std::endl;
  std::cout << "FINAL whole revolution fit" << std::endl;

  // Generate fit for the revolution
  Pass finalFit = rtxFile->fitRevolution(true);

  // SPRx24
  os << " " << std::endl;
  os << "...ended whole revolution fit" << std::endl;
  std::cout << " " << std::endl;
  std::cout << "...ended whole revolution fit" << std::endl;

  if ( (nSymCorrections + nSpikeCorrections) > 0 ){

    os << "Result for Revolution " 
       << Revolution
       << " Start: "
       << Start
       << " End: " 
       << End 
       << " Before: "
       << initialFit.getRMS()
       << " Middle: "
       << intermediateFit.getRMS()
       << " After: "
       << finalFit.getRMS()
       << " Improvement: "
       << ((initialFit.getRMS() - finalFit.getRMS()) / finalFit.getRMS()) * 100.0
       << "%"
       << std::endl;

  }
  else {
    os << "Result for Revolution " 
       << Revolution
       << " Start: "
       << Start
       << " End: " 
       << End 
       << " Before: "
       << initialFit.getRMS()
       << " Middle: "
       << intermediateFit.getRMS()
       << " After: "
       << finalFit.getRMS()
       << " Improvement: 0% (no corrections applied)"
       << std::endl;
  }

  if (finalFit.getRMS() > 0.001) {
    os << "WARNING: Revolution with high RMS. Revolution: "
       << Revolution
       << " RMS: "
       << finalFit.getRMS()
       << std::endl;
  }

  stringstream f2; f2 << finalFit.getRMS();
 
  // NEW CR x7
  os << "RTX final fit being plotted (SYMFIX)  ===" << std::endl;
  errstr << "RTX final fit being plotted (SYMFIX)  ===" << std::endl;
  std::cout << "RTX final fit being plotted (SYMFIX)  ===" << std::endl;

  rtxFile->plotRevolution("Fit of reconstructed data for rev " + revString + " RMS: " +  f2.str(),"", "_SYMFIX.ps");

  if (hasOBTreset) {
    	os << "WARNING: OBT Reset detected in revolution "
    	   << Revolution
    	   << std::endl;    
  }

  // SPRx31 Move the no fit data towards new defined columns
  os << "RTX moving nofit data ===" << std::endl;
  rtxFile->moveup();

  // Cleanup the dataset
  // SPRx28 cleanup activated
  // CRx22 cleanup may be not performed
  if (keepAncilliary) {
     os << "RTX will *not* clean up non ICD compliant columns ===" << std::endl;
  } else {
     os << "RTX cleaning up non ICD compliant columns ===" << std::endl;
     rtxFile->cleanup();
  } 
  // Close the dataset  
  rtxFile->close();

  // NEW CR x7
  os << "RTX final closing ===" << std::endl;
    
}



